/* Copyright 1998-2017 Research Foundation State University of New York  */

/* This file is part of QUB Online.                                      */

/* QUB Online is free software; you can redistribute it and/or modify   */
/* it under the terms of the GNU General Public License as published by  */
/* the Free Software Foundation, either version 3 of the License, or     */
/* (at your option) any later version.                                   */

/* QUB Online is distributed in the hope that it will be useful,        */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/* GNU General Public License for more details.                          */

/* You should have received a copy of the GNU General Public License,    */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see */
/* <http://www.gnu.org/licenses/>.                                       */

#include <math.h>
#include <stdlib.h>
#include "qubx_amp.h"
#include "asmffi.h"
#include "matrix.h"
#include "matrixutil.h"
#include "mbaum.h"

using std::endl;
using std::ostream;
using std::max;

class amp_model {
public:
  int nstate;
  int npath;
  int nclass;
  fq::vector<int> clazz;
  fq::vector<double> pr;
  fq::matrix<int> path;
  fq::vector<double> x, q, amp, xms;
  fq::vector<int> srcCls, destCls;
  fq::matrix<double> mtx;
  fq::vector<double> vct;

  amp_model(qubx_model *src)
    : nstate(src->Nstate)
  {
    // condense the classes--only occupied ones
    for (int i=0; i<nstate; ++i) {
      int c = src->clazz[i];
      int d = -1;
      for (int j=0; j<srcCls.size(); ++j) {
        if ( srcCls[j] == c ) {
          d = j;
          break;
        }
      }
      if ( d < 0 ) {
        d = srcCls.size();
        srcCls.push_back(c);
      }
      destCls.push_back(d);
    }

    for (int i=0; i<nstate; ++i) {
      clazz.push_back(destCls[src->clazz[i]]);
      pr.push_back(1.0/nstate);
    }

    nclass = srcCls.size();
    for (int i=0; i<nclass; ++i) {
      int c = srcCls[i];
      amp.push_back(src->amp[c]);
      xms.push_back(sqrt(src->var[c]));
    }

    fq::vector<int> pathRow(4);
    pathRow[2] = -1;
    pathRow[3] = 0;
    for (int a=0; a<nstate; ++a) {
      for (int b=a+1; b<nstate; ++b) {
        if ( src->K0[a][b] ) {
          pathRow[0] = a;
          pathRow[1] = b;
          path.push_back(pathRow);
          x.push_back( log(src->K0[a][b]) );
          x.push_back( src->K1[a][b] );
          pathRow[0] = b;
          pathRow[1] = a;
          path.push_back(pathRow);
          x.push_back( log(src->K0[b][a]) );
          x.push_back( src->K1[b][a] );
        }
      }
    }
    npath = path.nr;
    mtx.resize(2*npath+1, 2*npath+1);
    vct.resize(2*npath+1);
  }
};

// See also CChannelCombine in QubOpt for description 
void nexcom0(int iChannels, int iStates, int *paStates, int *piContinue, int *pTemp, int *piStateIndex) {
	if( *piContinue==0 ) {
		paStates[1] = iChannels;
        *pTemp = iChannels;
        *piStateIndex = 0;
        for( int iState=2; iState<=iStates; iState++) 
			paStates[iState]=0;
		}
	else {
		if(*pTemp > 1) 
			*piStateIndex=0;
		(*piStateIndex)++;
		*pTemp = paStates[*piStateIndex];
		paStates[*piStateIndex] = 0;
		paStates[1] = *pTemp - 1;
		paStates[*piStateIndex + 1] = paStates[*piStateIndex + 1] + 1;
		} 
    
	*piContinue=(paStates[iStates] != iChannels) ? 1 : 0 ; 
	}

/* DZERO is the largest number in double precision which when added  to one gives one. */  
/* FZERO is the largest number in single precision which when added  to one gives one. */      
// #define FZERO   1.0e-8
// For notes - see eclass description in QubOpt 
#define DZERO   1.0e-15 

void eclass(double *a, int n, int *nc, double *c, int *ia) {
    int  i,k,m,nl,*il,*ii;
    double  amin;
	double openDirection = (n > 1) ? (a[1] - a[0]) : 1.0;
	
    il = ialloc(n);
    ii = ialloc(n);
	
    for (i=0; i<n; i++) il[i]=i;
    nl=n;
    *nc=0;
    while (nl>0) {
		k=il[0];
		amin=a[k];
		for (i=1; i<nl; i++) {
			k=il[i];
			if ( ((openDirection > 0.0) && (a[k]-amin<DZERO)) || ((openDirection < 0.0) && (amin-a[k]<DZERO)) )
				amin=a[k];
			}
		c[*nc]=amin;
		for (m=i=0; i<nl; i++) {
			k=il[i]; 
			if (fabs(a[k]-amin)<DZERO) ia[k]=*nc;
			else {
				ii[m]=k;
				m++;
				}
			}
		for (i=0; i<m; i++) il[i]=ii[i];
		nl=m;
		(*nc)++;
		}   
	
    ifree(il);
    ifree(ii);
	}

void mutimakv(int nchannel, int nstate, int npath, 
			  int *state, int **path, double *amp, 
			  int *nmutistate, int *nmutipath, int *nmuticlass, 
			  int ***mutistate, int ***mutipath){
	int	i, k, l, m, n, s, t, iContinue, iStateIndex, *z, *ia, nn;
    double tmp, *a, *c;

	//model size for memory allocation
	tmp = 0.0;
    for( n=1; n<nstate; n++)
		tmp += log(double(n+nchannel)) - log(double(n));

	nn = (int)(exp(tmp) + 1);
    *mutistate = ialloc2(nn, nstate + 1);

	//muti-channel states
	int temp;
    z = ialloc(nstate + 1);
    *nmutistate = iContinue = 0; 
    do {
		nexcom0(nchannel, nstate, z, &iContinue, &temp, &iStateIndex);
		for (i = 0; i < nstate; i++)
			(*mutistate)[*nmutistate][i + 1] = z[i + 1];
		(*nmutistate)++;
		} while (iContinue == 1);
    ifree(z);
    
	//conductance classes
    a  = dalloc(*nmutistate);
    ia = ialloc(*nmutistate);
    c  = dalloc(*nmutistate);
        
    for (n = 0; n < *nmutistate; n++) {
		for (a[n] = 0., i = 0; i < nstate; i++) {
			k = state[i];
			a[n] += (*mutistate)[n][i + 1]*amp[k];
			}
		}

    eclass(a, *nmutistate, nmuticlass, c, ia);
    for( n=0; n<*nmutistate; n++ )
		(*mutistate)[n][0] = ia[n];
   
	//number of paths for memory allocation
	n=0;
	for( k=0; k<*nmutistate; k++ ) {
		for( l=0; l<*nmutistate; l++ ) {
			s=0;
			for( i=0; i<nstate; i++)
				s += abs((*mutistate)[k][i + 1] - (*mutistate)[l][i + 1]);
			if (s == 2)
				n++;
			}
		}
    
	*mutipath = ialloc2(n, 4); 
    
	//transition paths        
    for (*nmutipath = m = 0; m < *nmutistate; m++) {
		for( n=0; n<*nmutistate; n++) {
			s=0;
			for( i=0; i<nstate; i++)
				s += abs((*mutistate)[m][i + 1] - (*mutistate)[n][i + 1]);
			
			if (s == 2) {
				for( i=0; i<nstate; i++) {
					t = (*mutistate)[m][i + 1] - (*mutistate)[n][i + 1];
					if (t == 1) 
						k = i;
					if (t == -1) 
						l = i;
					}	
				
				for( i=0; i<npath; i++) {
					if (path[i][0] == k && path[i][1] == l) {
						(*mutipath)[*nmutipath][0] = m;
						(*mutipath)[*nmutipath][1] = n;
						(*mutipath)[*nmutipath][2] = k;
						(*mutipath)[*nmutipath][3] = l;
						(*nmutipath)++;
						}
					}
				}
			}
		}
    
    dfree(a);
    ifree(ia);  
    dfree(c);
	}

void metamakv(int nstate, int namp, int **state, int nar, int *nmetastate, int ***metastate){
    int  I, J, i, j, k, l, n, s;
    
	//model size for memory allocation 
    k = nstate*(int)pow((double)namp, nar);   
    l = (nar + 3 + nstate*(namp + 1)) + 1; // +1 because it wasn't enough!
    *metastate = ialloc2(k, l);

	//meta-states      
	*nmetastate = 0;
    for( j=0; j<nstate; j++) {
		(*metastate)[*nmetastate][0] = j;
		for( k=1; k<=nar; k++)
			(*metastate)[*nmetastate][k] = j;
		(*nmetastate)++;
		}   

    n = *nmetastate;
    for( l=1; l<=nar; l++ ) {
		for( i=0; i<*nmetastate; i++ ) {
			for( j=1; j<namp; j++ ) {
				for( k=0; k<=nar; k++ ) {
					if (k == l)
						(*metastate)[n][k] = j;
					else
						(*metastate)[n][k]=(*metastate)[i][k];
					}
				n++;
				}
			}
		*nmetastate = n;
    	}

	//set the 2nd column to be conductance index
	for( n=0; n<*nmetastate; n++ ) {
		for( k=nar+1; k>=1; k-- )
			(*metastate)[n][k] = (*metastate)[n][k-1];
		i = (*metastate)[n][1];
		(*metastate)[n][1] = state[i][0];
		}

	//allowable transitions leaving meta-state I 
    for( I=0; I<*nmetastate; I++ ) {
		n=0;
		for( J=0; J<*nmetastate; J++) {
			s=0;
			for( k=2; k<=nar+1; k++)	// TODO - instead test expr. and break since s is not needed.
				s += abs((*metastate)[J][k] - (*metastate)[I][k - 1]);
			if( s==0 )
				(*metastate)[I][nar+3+(n++)] = J;
			}
		(*metastate)[I][nar + 2] = n;
    	}

	//allowable transitions entering meta-state J
    for( J=0; J<*nmetastate; J++ ) {
		n=0;
		for( I=0; I<*nmetastate; I++ ) {
			s=0;
			for( k=2; k<=nar+1; k++)
				s += abs((*metastate)[J][k] - (*metastate)[I][k - 1]);
			if (s == 0)
				(*metastate)[J][nar+4+nstate+(n++)] = I;
			}
		(*metastate)[J][nar+3+nstate] = n;
		}
	}


/*
 * Constraint routine
 */
int freePar(int mc, int nx, double *x, double **Mtx, double *vct, int *nz, double *z, ostream &msgOut){
	const double TOL = 1.0e-5;
	const double MAD_SMALL = 1.0e-9;
	int     i,j,k,t; 
	double  **a,*b,**u,**v,*w,*h,wmax,s;
	
	/* no constraint */
	if (mc==0) {
		*nz=nx;
		for (i=0; i<nx; i++) {
			for (j=0; j<*nz; j++) 
				Mtx[i][j]=0.0;
			Mtx[i][i]=1.0;
			vct[i]=0.0;
			z[i]=x[i];
			}
		return TRUE;
		}
	
	/* initialization */
	a = dalloc2(mc+1,nx+1);
	b = dalloc(nx+1);
	u = dalloc2(mc+1,nx+1);
	v = dalloc2(nx+1,nx+1);
	w = dalloc(nx+1);
	h = dalloc(mc);
	
	for (i=0; i<mc; i++) {
		b[i]=vct[i];
		for (j=0; j<nx; j++)
			a[i][j]=Mtx[i][j];
		}
	
	/* check if initial guess meets constraints */
	for (i=0; i<mc; i++) {
		s=0.0;
		for (j=0; j<nx; j++) 
			s=s+a[i][j]*x[j];
		if(fabs(s-b[i])>1.0e-2) {
			msgOut << "Fatal err: Initial guess doesn't meet constraints: " << (s-b[i]) << endl;
			return FALSE;
			}
		}
	
	/* SVD */
	for (i=0; i<mc; i++) 
		for (j=0; j<nx; j++) 
			u[i][j]=a[i][j];
	
	//-----------------------------------------------------------------------
	//----- SVD 
	svdecomp0(u,mc,nx,w,v);

	//----- Set wmax = max(w[j=0..nx]).  Our version of svdcmp
	// performs a sort at the end so this will always be w[0].
	wmax=fabs(w[0]);
	for( j=1; j<nx; j++)
		if(fabs(w[j])>wmax) 
			wmax=fabs(w[j]);

	t=0; 
	while(fabs(w[t])>wmax*TOL) 
		t++;
	
	/* confirm SVD results */   
	if ( t != mc ) {
		msgOut << "Too many constraints.  " << (mc - t) << " constraints may be linear combinations of the others." << endl;
		// msgOut << "Fatal err: Bad SVD in freepar" << endl;
		return FALSE;
		}

	for (i=t; i<nx; i++) {
		if(fabs(w[i])>wmax*TOL){ // w[t] changed to w[i] Chris 6/2/2
			msgOut << "Fatal err: Bad SVD in freepar" << endl;
			return FALSE;
			}
		}

	for (i=0; i<mc; i++) {
		for (j=mc; j<nx; j++) {
			if(fabs(u[i][j])>TOL) {
				msgOut << "Fatal err: Bad SVD in freepar" << endl;
				return FALSE;
				}
			}
		}

	for (i=0; i<mc; i++) { // make sure that u's columns are orthogonal unit vectors
		for (j=0; j<mc; j++) { // changed from nx to mc Chris 6/2/2
			s=0.0;
			for (k=0; k<mc; k++) 
				s=s+u[k][i]*u[k][j];
			if( (i!=j&&fabs(s)>TOL) || (i==j&&fabs(s-1.0)>TOL) ) {
				msgOut << "Fatal err: Bad SVD in freepar" << endl;
				return FALSE;
				}
			}
		}
	for (i=0; i<nx; i++) { // make sure that v's columns are orthogonal unit vectors
		for (j=0; j<nx; j++) {
			s=0.0;
			for (k=0; k<nx; k++) 
				s=s+v[k][i]*v[k][j];
			if( (i!=j && fabs(s)>TOL) || (i==j && fabs(s-1.0)>TOL) ){
				msgOut << "Fatal err: Bad SVD in freepar" << endl;
				return FALSE;
				}
			}
		}
	for (i=0; i<nx; i++) {
		for (j=0; j<nx; j++) {
			s=0.0;
			for (k=0; k<nx; k++) 
				s=s+v[i][k]*v[j][k];
			if( (i!=j&&fabs(s)>TOL) || (i==j&&fabs(s-1.0)>TOL) ){
				msgOut << "Fatal err: Bad SVD in freepar" << endl;
				return FALSE;
				}
			}
		}  
	
	/* set Mtx and vct */
	for (i=0; i<mc; i++) {
		h[i]=0.0;
		for (j=0; j<mc; j++) 
			h[i]=h[i]+u[j][i]*b[j];
		if (i < t) 
			h[i]=h[i]/w[i]; // i always less than t anyway?
		if (i >= t && fabs(h[i])>TOL){
			msgOut << "Fatal err: Bad SVD in freepar" << endl;
			return FALSE;
			}
		}
	
	for (i=0; i<nx; i++) {
		vct[i]=0.0;
		for (j=0; j<t; j++) 
			vct[i]=vct[i]+v[i][j]*h[j];
		}
	
	for (i=0; i<nx; i++) {
		for (j=t; j<nx; j++) {
			if ( fabs(v[i][j]) < MAD_SMALL ) // 5/31/2 Chris:  some weird svd was strewing 1.0e-18 when using loop balance
				Mtx[i][j-t] = 0.0;
			else
				Mtx[i][j-t]=v[i][j];
			}
		}
		
	*nz=nx-mc;
	
	/* get z from x */
	for (i=0; i<*nz; i++) {
		z[i]=0.0;
		for (j=0; j<nx; j++)
			z[i]=z[i]+Mtx[j][i]*x[j];
		}

	// TESTING: scale z to start at 1.0 (unless it's zero, stay at zero)
	// effectively, define diagonal matrix S with entries from z (or 1 instead of 0)
	// so x = Mtx * S * 1
	// Then (Mtx*S) can replace Mtx, and the params start as an all 1 vector (except any zeros remain)
	for (i=0; i<*nz; ++i) {
		if ( fabs(z[i]) > MAD_SMALL ) {
			for (j=0; j<nx; ++j)
				Mtx[j][i] *= z[i];
			z[i] = 1.0;
		}
	}
	
	dfree2(a);
	dfree(b);
	dfree2(v);
	dfree2(u);
	dfree(w);
	dfree(h);
	return TRUE;
	}

// dx = a
// x = b + az	(matrix multiply)
void amp_ztox(int nz, int nx, double **a, double *b, double *z, double *x, double **dx){
    int  i, j;
	double xsum;
                  
    for( i=0; i<nx; i++ ) {
		xsum = b[i];
		for( j=0; j<nz; j++ ) 
			xsum += (dx[i][j]=a[i][j]) * z[j];
		x[i]=xsum;
		}
	}
// rate constants Q[0..nstate][0..nstate] and dq[nstate][nstate][nz]
void amp_xtoq(int nstate, int npath, int **path, double *C, 
					double V, double *x, double **q, int nz, double **dx, double ***dq) {
    int i, j, k, m;
	double nQ;
     
	//First zero the Q and dq matrices
    for( i=0; i<nstate; i++) {
		for( j=0; j<nstate; j++) {
			q[i][j] = 0.0;
		    for( k=0; k<nz; k++) 
				dq[i][j][k] = 0.0;
			}
		}

	//Calculate the rates for the connected states and the diagonal
	for( m=0; m<npath; m++ ) {
		i = path[m][0];	//from state 
		j = path[m][1];	//to state
		k = path[m][2];	//drug index
		nQ = ((k==-1) ? 1 : C[k]) * exp(x[2*m] + V*x[2*m + 1]);

		//the q matrix - the rate expression.  diagonal = -sum(row)
		q[i][i] -=  ( q[i][j] = nQ );	
		// the dq matrix 
		for( k=0; k<nz; k++)
			dq[i][i][k] -=  ( dq[i][j][k] = nQ *(dx[2*m][k] + V*dx[2*m + 1][k]) );
		}
	}

void amp_qtomutiq(int nmutistate,
			  int **mutistate, int nmutipath, int **mutipath, double **q,
			  double **qq, int nz, double ***dq, double ***dqq){   
    int	I, J, i, j, k, m;
    double s;                
    
 	//Rate constants
    for( I=0; I<nmutistate; I++)
		for( J=0; J<nmutistate; J++)
			qq[I][J]=0.0;

    for( m=0; m<nmutipath; m++) {
		I = mutipath[m][0];
		J = mutipath[m][1];
		i = mutipath[m][2];
		j = mutipath[m][3];
		qq[I][J] = mutistate[I][i + 1]*q[i][j];
		}

    for( I=0; I<nmutistate; I++) {
		s = 0.0;
		for( J=0; J<nmutistate; J++)
			if (J != I)
				s -= qq[I][J];
		qq[I][I] = s;
		}
    

	//derivatives to z-variables
    for( I=0; I<nmutistate; I++ )
		for( J=0; J<nmutistate; J++ )
			for( k=0; k<nz; k++ )
				dqq[I][J][k] = 0.0;
    
    for( m=0; m<nmutipath; m++ ) {
		I = mutipath[m][0];
		J = mutipath[m][1];
		i = mutipath[m][2];
		j = mutipath[m][3];
		for( k=0; k<nz; k++ )
			dqq[I][J][k] = mutistate[I][i + 1]*dq[i][j][k];
		} 
    
    for( k=0; k<nz; k++ ) {
		for( I=0; I<nmutistate; I++) {
			s = 0.0;
			for( J=0; J<nmutistate; J++) 
				if (J != I)
					s -= dqq[I][J][k];
			dqq[I][I][k] = s;
			}
		}
	}

/*
 * Calculate A=I*t^i0+Qt^(i0+1)/(i0+1)+Q^2*t^(i0+2)/(i0+2)+... 
 * and its derivatives.
 */

  
int nterm0(double **q, int n, double t){      
    int  i,j,m;
    double s=0.0,eps=1.0e-50;
           
    for (i=0; i<n; i++)   
	    for (j=0; j<n; j++) 
			s += t*fabs(q[i][j]);
	
	double r=1.0;
    for (m=0; r/(1.0-r/(m+2))>=eps; m++) 
		r*=s/(m+1);
	// Note : Qubopt version returns max(m,50) - only difference from this 
    return max(m, 30);
	}


#define EXIT_ERROR /*fprintf(stderr,"Memory allocation error\n"); exit(1);*/return NULL;

double  *** dalloc3(int row, int col, int depth) {
    int i,j;
    double *** ptr = (double ***)malloc(row*sizeof(double **));
    if (ptr==NULL)
       EXIT_ERROR
    
    ptr[0] = (double **)malloc(row*col*sizeof(double *));
    if (ptr[0]==NULL)
       EXIT_ERROR
    
    for (i=1; i<row; i++) 
		ptr[i]=&(ptr[0][i*col]);
    
    ptr[0][0] = (double *)malloc(row*col*depth*sizeof(double));
    if (ptr[0][0]==NULL)
       EXIT_ERROR
    
    for (i=0; i<row; i++)
		for (j=0; j<col; j++) 
			ptr[i][j]=&(ptr[0][0][(i*col+j)*depth]);
    return ptr;
	}

void free3(char ***ptr) {
    free(ptr[0][0]);   
    free(ptr[0]);
    free(ptr);
	}

#define dirac(a,b)  ((a)==(b) ? 1 : 0)

void amp_mexp(int n, double t, double **q, double **a, int nz, 
              double ***dq, double ***da, int i0) {
    int  i,j,k,l,it,itmax;
    double  s,**u,**u1,***v,***v1;
    
    u  = dalloc2(n,n);
    u1 = dalloc2(n,n);
    v  = dalloc3(n,n,nz);
    v1 = dalloc3(n,n,nz);
    
	for (i=0; i<n; i++)    
		for (j=0; j<n; j++) {
			a[i][j]=0.0;
			u[i][j]=dirac(i,j)*pow(t,i0);
			for (k=0; k<nz; k++) 
				da[i][j][k]=v[i][j][k]=0.0;
			}
    
    itmax = nterm0(q,n,t);   
        
	for (it=0; it<=itmax; it++) {
		for (i=0; i<n; i++)         
			for (j=0; j<n; j++) {
				a[i][j] += u[i][j];
				for (k=0; k<nz; k++) 
					da[i][j][k]+=v[i][j][k];
				}
		for (i=0; i<n; i++)
			for (j=0; j<n; j++) { 
				for (s=0.,l=0; l<n; l++) 
					s+=u[i][l]*q[l][j]; 
				u1[i][j]=s; 
				} 
		for (k=0; k<nz; k++)  
			for (i=0; i<n; i++)         
				for (j=0; j<n; j++) {
					for (s=0.,l=0; l<n; l++) 
						s+=q[i][l]*v[l][j][k]+dq[i][l][k]*u[l][j];
					v1[i][j][k]=s;
					}
		for (i=0; i<n; i++)         
			for (j=0; j<n; j++) {
				u[i][j]=u1[i][j]*t/(it+1+i0);
				for (k=0; k<nz; k++) 
					v[i][j][k]=v1[i][j][k]*t/(it+1+i0);
				}
		}
        
	dfree2(u);
	dfree2(u1);
	free3((char***)v);
	free3((char***)v1);   
}

void qubx_amp_autoinit(int n, float *data, amp_model& model, int autoInit, int autoInitUp) {
  if ( autoInit && n && (model.nclass > 1) ) {
    float lo = data[0];
    float hi = lo;
    for (int i=1; i<n; ++i) {
      float x = data[i];
      if ( x < lo )
        lo = x;
      if ( x > hi )
        hi = x;
    }
    float d = (hi - lo) / (float) (model.nclass - 1);
    if ( ! autoInitUp ) {
      float t = lo;
      lo = hi;
      hi = t;
    }
    for (int i=0; i<model.nclass; ++i) {
      model.amp[i] = lo + ((float)i)*d;
      model.xms[i] = .2*d;
    }
  }
}  

int qubx_amp_makeEvents(int *classes, int *counts, amp_model& model, int *stateOut, int n) {
  int Nevent = 0;
  if ( classes && n ) {
    int c = model.clazz[stateOut[0]];
    int count = 1;
    for (int i=1; i<n; ++i) {
      int d = model.clazz[stateOut[i]];
      if ( d == c ) {
        ++count;
      } else {
        classes[Nevent] = model.srcCls[c];
        counts[Nevent] = count;
        ++Nevent;
        c = d;
        count = 1;
      }
    }
    classes[Nevent] = model.srcCls[c];
    counts[Nevent] = count;
    ++Nevent;
  }
  return Nevent;
}

void discardMBaumOutput(char * msg) {
}

extern "C" int QUBFAST_API qubx_amp(double sampling, int n, float *data, qubx_model *_model, int autoInit, int autoInitUp, int *classes, int *counts) {
  amp_model model(_model);
  qubx_amp_autoinit(n, data, model, autoInit, autoInitUp);

  fq::vector<double> C(2);
  C[0] = 0.0;
  C[1] = 1.0;
  int nfir = 0;
  fq::vector<double> fir(1);
  fir[0] = 1.0;

  fq::vector<int> idrug(model.npath);
  for (int i=0; i<model.npath; ++i) {
    idrug[i] = 0;
  }
  fq::vector<double> ratio(model.nclass);
  for (int i=0; i<model.nclass; ++i) {
    ratio[i] = model.amp[i] - model.amp[0];
  }
  int nmutistate, nmutipath, nmuticlass;
  int **mutistate, **mutipath;
  mutimakv(1, model.nstate, model.npath, model.clazz, model.path, ratio,
           &nmutistate, &nmutipath, &nmuticlass, &mutistate, &mutipath);

  int nmetastate = 0;
  int **metastate = 0;
  metamakv(nmutistate, nmuticlass, mutistate, 0, &nmetastate, &metastate);

  int nz = 2*model.npath;
  fq::vector<double> z(nz);
  freePar(0, nz, model.x, model.mtx, model.vct, &nz, z, console);
  z.resize(nz);

  fq::matrix<double> q(model.nstate, model.nstate), qq(nmetastate, nmetastate), a(nmetastate+1, nmetastate+1), dx(model.x.size(), nz), dfa(nmetastate, nmetastate);
  fq::tensor<double> dq(model.nstate, model.nstate, nz), dqq(nmetastate, nmetastate, nz), da(nmetastate, nmetastate, nz);
  fq::matrix<double*> dqAsPtr(model.nstate, model.nstate), dqqAsPtr(nmetastate, nmetastate), daAsPtr(nmetastate, nmetastate);
  for (int i=0; i<model.nstate; ++i) {
    for (int j=0; j<model.nstate; ++j) {
      dqAsPtr[i][j] = & dq[i][j][0];
    }
  }
  for (int i=0; i<nmetastate; ++i) {
    for (int j=0; j<nmetastate; ++j) {
      dqqAsPtr[i][j] = & dqq[i][j][0];
    }
  }
  for (int i=0; i<nmetastate; ++i) {
    for (int j=0; j<nmetastate; ++j) {
      daAsPtr[i][j] = & da[i][j][0];
    }
  }

  amp_ztox(z.size(), 2*model.npath, model.mtx, model.vct, z, model.x, dx);
  amp_xtoq(model.nstate, model.npath, model.path, C, 0.0, model.x, q, z.size(), dx, dqAsPtr);
  amp_qtomutiq(nmutistate, mutistate, nmutipath, mutipath, q, qq, z.size(), dqAsPtr, dqqAsPtr);
  amp_mexp(nmutistate, (float) sampling, qq, a, z.size(), dqqAsPtr, daAsPtr, 0);

  int nar = 0;
  fq::matrix<double> ar(nmuticlass, 1);
  for (int i=0; i<nmuticlass; ++i) {
    double xms = model.xms[i];
    ar[i][0] = xms*xms;
  }

  double ll = 0.0;
  fq::vector<int> stateOut(n);
  int processFlag = 2;
  mbaum(n, data, nmutistate, a, nmetastate, metastate, model.pr, nmuticlass, model.amp, model.xms, nar, ar, nfir, fir, &ll, stateOut, discardMBaumOutput, &processFlag);

  for (int i=0; i<model.nclass; ++i) {
    _model->amp[model.srcCls[i]] = model.amp[i];
    _model->var[model.srcCls[i]] = model.xms[i] * model.xms[i];
  }

  int Nevent = qubx_amp_makeEvents(classes, counts, model, stateOut, n);
  
  ifree2(mutistate);
  ifree2(mutipath);

  return Nevent;
}
